import click
from oogway.key import Key
from oogway.validate import validate as validation

@click.command()
@click.option('--address', default='bech32', help="address format (Bech32 | P2SH | P2PKH)")
@click.option('--testnet', default=False, help="generate testnet keypair (True | False)")
def keypair(address, testnet):
    """generates keypair with specific address format"""
    key = Key(testnet=testnet, mnemonic_strength=128)
       
    address = address.lower()

    if address == 'bech32':
        _address = key.address(address)
    elif address == 'p2sh':
        _address = key.address(address)
    elif address == 'p2pkh':
        _address = key.address(address)
    else:
        raise click.BadParameter('Invalid address format passed', param_hint=["--address"])
    
    click.echo('\n------------------------------------')
    click.echo('MNEMONIC: %s' % key.mnemonic)
    click.echo('WIF: %s' % key.wif)
    click.echo('ADDRESS: %s' % _address)
    click.echo('------------------------------------\n')

@click.command()
@click.argument('addresses', nargs=-1)
def validate(addresses):
    """validate a set of addresses and returns format"""
    for address in addresses:
        v = validation.is_valid_address(address)
        try:
            _format = validation.address_format(address)
        except ValueError:
            _format = "INVALID"
        if v == True:
            click.echo('\n[VALID | %s] %s\n' % (_format, address))
        elif v == False:
            click.echo('\n[INVALID] %s\n' % address)



"""
Copyright (c) 2016 Ofek Lev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from collections import namedtuple
from itertools import islice
import math
import re
import logging
from random import randint, shuffle
from .utils.crypto import double_sha256, sha256
from .exceptions import InsufficientFunds
from .utils.format import address_to_public_key_hash, segwit_scriptpubkey
from .utils.utils import (
    bytes_to_hex,
    chunk_data,
    hex_to_bytes,
    int_to_unknown_bytes,
    int_to_varint,
    script_push,
    get_signatures_from_script,
    read_bytes,
    read_var_int,
    read_var_string,
    read_segwit_string,
)
from .utils.format import verify_sig, get_version
from .utils.base58 import b58decode_check
from .utils.bech32 import decode as segwit_decode
from .utils.constants import (
    TEST_SCRIPT_HASH,
    MAIN_SCRIPT_HASH,
    TEST_PUBKEY_HASH,
    MAIN_PUBKEY_HASH,
    VERSION_1,
    MARKER,
    FLAG,
    SEQUENCE,
    LOCK_TIME,
    HASH_TYPE,
    MESSAGE_LIMIT
)
from .utils.opcodes import (
    OP_0,
    OP_CHECKLOCKTIMEVERIFY,
    OP_CHECKSIG,
    OP_DUP,
    OP_EQUALVERIFY,
    OP_HASH160,
    OP_PUSH_20,
    OP_RETURN,
    OP_EQUAL
)

class TxIn:
    __slots__ = ('script_sig', 'script_sig_len', 'txid', 'txindex', 'witness', 'amount', 'sequence', 'segwit_input')

    def __init__(self, script_sig, txid, txindex, witness=b'', amount=None, sequence=SEQUENCE, segwit_input=False):

        self.script_sig = script_sig
        self.script_sig_len = int_to_varint(len(script_sig))
        self.txid = txid
        self.txindex = txindex
        self.witness = witness
        self.amount = amount
        self.sequence = sequence
        self.segwit_input = segwit_input

    def __eq__(self, other):
        return (
            self.script_sig == other.script_sig
            and self.script_sig_len == other.script_sig_len
            and self.txid == other.txid
            and self.txindex == other.txindex
            and self.witness == other.witness
            and self.amount == other.amount
            and self.sequence == other.sequence
            and self.segwit_input == other.segwit_input
        )

    def __repr__(self):
        if self.is_segwit():
            return 'TxIn({}, {}, {}, {}, {}, {}, {})'.format(
                repr(self.script_sig),
                repr(self.script_sig_len),
                repr(self.txid),
                repr(self.txindex),
                repr(self.witness),
                repr(self.amount),
                repr(self.sequence),
            )
        return 'TxIn({}, {}, {}, {}, {})'.format(
            repr(self.script_sig), repr(self.script_sig_len), repr(self.txid), repr(self.txindex), repr(self.sequence)
        )

    def __bytes__(self):
        return b''.join([self.txid, self.txindex, self.script_sig_len, self.script_sig, self.sequence])

    def is_segwit(self):
        return self.segwit_input or self.witness


Output = namedtuple('Output', ('address', 'amount'))


class TxOut:
    __slots__ = ('amount', 'script_pubkey_len', 'script_pubkey')

    def __init__(self, amount, script_pubkey):
        self.amount = amount
        self.script_pubkey = script_pubkey
        self.script_pubkey_len = int_to_varint(len(script_pubkey))

    def __eq__(self, other):
        return (
            self.amount == other.amount
            and self.script_pubkey == other.script_pubkey
            and self.script_pubkey_len == other.script_pubkey_len
        )

    def __repr__(self):
        return 'TxOut({}, {}, {})'.format(repr(self.amount), repr(self.script_pubkey), repr(self.script_pubkey_len))

    def __bytes__(self):
        return b''.join([self.amount, self.script_pubkey_len, self.script_pubkey])


class TxObj:
    __slots__ = ('version', 'TxIn', 'TxOut', 'locktime')

    def __init__(self, version, TxIn, TxOut, locktime):
        segwit_tx = any([i.segwit_input or i.witness for i in TxIn])
        self.version = version
        self.TxIn = TxIn
        if segwit_tx:
            for i in self.TxIn:
                i.witness = i.witness if i.witness else b'\x00'
        self.TxOut = TxOut
        self.locktime = locktime

    def __eq__(self, other):
        return (
            self.version == other.version
            and self.TxIn == other.TxIn
            and self.TxOut == other.TxOut
            and self.locktime == other.locktime
        )

    def __repr__(self):
        return 'TxObj({}, {}, {}, {})'.format(
            repr(self.version), repr(self.TxIn), repr(self.TxOut), repr(self.locktime)
        )

    def __bytes__(self):
        inp = int_to_varint(len(self.TxIn)) + b''.join(map(bytes, self.TxIn))
        out = int_to_varint(len(self.TxOut)) + b''.join(map(bytes, self.TxOut))
        wit = b''.join([w.witness for w in self.TxIn])
        return b''.join([self.version, MARKER if wit else b'', FLAG if wit else b'', inp, out, wit, self.locktime])

    def legacy_repr(self):
        inp = int_to_varint(len(self.TxIn)) + b''.join(map(bytes, self.TxIn))
        out = int_to_varint(len(self.TxOut)) + b''.join(map(bytes, self.TxOut))
        return b''.join([self.version, inp, out, self.locktime])

    def to_hex(self):
        return bytes_to_hex(bytes(self))

    @classmethod
    def is_segwit(cls, tx):
        if isinstance(tx, cls):
            tx = bytes(tx)
        elif not isinstance(tx, bytes):
            tx = hex_to_bytes(tx)
        return tx[4:6] == MARKER + FLAG


def calc_txid(tx_hex):
    tx_obj = deserialize(tx_hex)
    return bytes_to_hex(double_sha256(tx_obj.legacy_repr())[::-1])


def estimate_tx_fee(in_size, n_in, out_size, n_out, satoshis, segwit=False):

    if not satoshis:
        return 0

    estimated_size = math.ceil(
        in_size
        + len(int_to_unknown_bytes(n_in, byteorder='little'))
        + out_size
        + len(int_to_unknown_bytes(n_out, byteorder='little'))
        + 8
        # Accounting for magic header vBytes ('0001') 
        + (0.5 if segwit else 0)
    )

    estimated_fee = estimated_size * satoshis

    logging.debug('Estimated fee: {} satoshis for {} bytes'.format(estimated_fee, estimated_size))

    return estimated_fee


def select_coins(target, fee, output_size, min_change, *, absolute_fee=False, consolidate=False, unspents):
    '''
    Implementation of Branch-and-Bound coin selection defined in Erhart's
    Master's thesis An Evaluation of Coin Selection Strategies here:
    http://murch.one/wp-content/uploads/2016/11/erhardt2016coinselection.pdf

    :param target: The total amount of the outputs in a transaction for which
                   we try to select the inputs to spend.
    :type target: ``int``
    :param fee: The number of satoshi per byte for the fee of the transaction.
    :type fee: ``int``
    :param output_size: A list containing as int the sizes of each output.
    :type output_size: ``list`` of ``int`
    :param min_change: The minimum amount of satoshis allowed for the
                       return/change address if there is no perfect match.
    :type min_change: ``int``
    :param absolute_fee: Whether or not the parameter ``fee`` should be
                         repurposed to denote the exact fee amount.
    :type absolute_fee: ``bool``
    :param consolidate: Whether or not the Branch-and-Bound process for finding
                        a perfect match should be skipped and all unspents
                        used directly.
    :type consolidate: ``bool``
    :param unspents: The UTXOs to use as inputs.
    :type unspents: ``list`` of :class:`~bit.network.meta.Unspent`
    :raises InsufficientFunds: If ``unspents`` does not contain enough balance
                               to allow spending matching the target.
    '''

    # The maximum number of tries for Branch-and-Bound:
    BNB_TRIES = 1000000

    # COST_OF_OVERHEAD excludes the return address of output_size (last element).
    COST_OF_OVERHEAD = (8 + sum(output_size[:-1]) + 1) * fee

    def branch_and_bound(d, selected_coins, effective_value, target, fee, sorted_unspents):  # pragma: no cover

        nonlocal COST_OF_OVERHEAD, BNB_TRIES
        BNB_TRIES -= 1
        COST_PER_INPUT = 148 * fee  # Just typical estimate values
        COST_PER_OUTPUT = 34 * fee

        # The target we want to match includes cost of overhead for transaction
        target_to_match = target + COST_OF_OVERHEAD
        # Allowing to pay fee for a whole input and output is rationally
        # correct, but increases the fee-rate dramatically for only few inputs.
        match_range = COST_PER_INPUT + COST_PER_OUTPUT
        # We could allow to spend up to X% more on the fees if we can find a
        # perfect match:
        # match_range += int(0.1 * fee * sum(u.vsize for u in selected_coins))

        # Check for solution and cut criteria:
        if effective_value > target_to_match + match_range:
            return []
        elif effective_value >= target_to_match:
            return selected_coins
        elif BNB_TRIES <= 0:
            return []
        elif d >= len(sorted_unspents):
            return []
        else:
            # Randomly explore next branch:
            binary_random = randint(0, 1)
            if binary_random:
                # Explore inclusion branch first, else omission branch:
                effective_value_new = effective_value + sorted_unspents[d].amount - fee * sorted_unspents[d].vsize

                with_this = branch_and_bound(
                    d + 1, selected_coins + [sorted_unspents[d]], effective_value_new, target, fee, sorted_unspents
                )

                if with_this != []:
                    return with_this
                else:
                    without_this = branch_and_bound(
                        d + 1, selected_coins, effective_value, target, fee, sorted_unspents
                    )

                    return without_this

            else:
                # As above but explore omission branch first:
                without_this = branch_and_bound(d + 1, selected_coins, effective_value, target, fee, sorted_unspents)

                if without_this != []:
                    return without_this
                else:
                    effective_value_new = effective_value + sorted_unspents[d].amount - fee * sorted_unspents[d].vsize

                    with_this = branch_and_bound(
                        d + 1, selected_coins + [sorted_unspents[d]], effective_value_new, target, fee, sorted_unspents
                    )

                    return with_this

    sorted_unspents = sorted(unspents, key=lambda u: u.amount, reverse=True)
    selected_coins = []

    if not consolidate and not absolute_fee:
        # Trying to find a perfect match using Branch-and-Bound:
        selected_coins = branch_and_bound(
            d=0, selected_coins=[], effective_value=0, target=target, fee=fee, sorted_unspents=sorted_unspents
        )
        remaining = 0

    # Fallback: If no match, Single Random Draw with return address:
    if selected_coins == []:
        unspents = unspents.copy()
        # Since we have no information on the user's spending habit it is
        # best practice to randomly select UTXOs until we have enough.
        if not consolidate:
            # To have a deterministic way of inserting inputs when
            # consolidating, we only shuffle the unspents otherwise.
            shuffle(unspents)
        while unspents:
            selected_coins.append(unspents.pop(0))
            estimated_fee = estimate_tx_fee(
                sum(u.vsize for u in selected_coins), len(selected_coins), sum(output_size), len(output_size), fee, any(u.segwit for u in selected_coins)
            )
            estimated_fee = fee if absolute_fee else estimated_fee
            remaining = sum(u.amount for u in selected_coins) - target - estimated_fee
            if remaining >= min_change and (not consolidate or len(unspents) == 0):
                break
        else:
            raise InsufficientFunds(
                'Balance {} is less than {} (including '
                'fee).'.format(sum(u.amount for u in selected_coins), target + min_change + estimated_fee)
            )

    return selected_coins, remaining


def deserialize(tx):
    if isinstance(tx, str) and re.match('^[0-9a-fA-F]*$', tx):
        return deserialize(hex_to_bytes(tx))

    segwit_tx = TxObj.is_segwit(tx)

    version, tx = read_bytes(tx, 4)

    if segwit_tx:
        _, tx = read_bytes(tx, 1)  # ``marker`` is nulled
        _, tx = read_bytes(tx, 1)  # ``flag`` is nulled

    ins, tx = read_var_int(tx)
    inputs = []
    for i in range(ins):
        txid, tx = read_bytes(tx, 32)
        txindex, tx = read_bytes(tx, 4)
        script_sig, tx = read_var_string(tx)
        sequence, tx = read_bytes(tx, 4)
        inputs.append(TxIn(script_sig, txid, txindex, sequence=sequence))

    outs, tx = read_var_int(tx)
    outputs = []
    for _ in range(outs):
        amount, tx = read_bytes(tx, 8)
        script_pubkey, tx = read_var_string(tx)
        outputs.append(TxOut(amount, script_pubkey))

    if segwit_tx:
        for i in range(ins):
            wnum, tx = read_var_int(tx)
            witness = int_to_varint(wnum)
            for _ in range(wnum):
                w, tx = read_segwit_string(tx)
                witness += w
            inputs[i].witness = witness

    locktime, _ = read_bytes(tx, 4)

    txobj = TxObj(version, inputs, outputs, locktime)

    return txobj


def sanitize_tx_data(
    unspents,
    outputs,
    fee,
    leftover,
    combine=True,
    message=None,
    compressed=True,
    absolute_fee=False,
    min_change=0,
    version='main',
    message_is_hex=False,
    replace_by_fee=False
):
    """
    sanitize_tx_data()

    fee is in satoshis per byte.
    """

    outputs = outputs.copy()

    for i, output in enumerate(outputs):
        dest, amount = output
        outputs[i] = (dest, amount)

    if not unspents:
        raise ValueError('Transactions must have at least one unspent.')

    # Temporary storage so all outputs precede messages.
    messages = []

    if message:
        if message_is_hex:
            message_chunks = chunk_data(message, MESSAGE_LIMIT)
        else:
            message_chunks = chunk_data(message.encode('utf-8'), MESSAGE_LIMIT)

        for message in message_chunks:
            messages.append((message, 0))

    # Include return address in output count.
    # Calculate output size as a list (including return address).
    output_size = [len(address_to_scriptpubkey(o[0])) + 9 for o in outputs]
    output_size.append(len(messages) * (MESSAGE_LIMIT + 9))
    output_size.append(len(address_to_scriptpubkey(leftover)) + 9)
    sum_outputs = sum(out[1] for out in outputs)

    # Use Branch-and-Bound for coin selection:
    unspents[:], remaining = select_coins(
        sum_outputs,
        fee,
        output_size,
        min_change=min_change,
        absolute_fee=absolute_fee,
        consolidate=combine,
        unspents=unspents,
    )

    if replace_by_fee:
        for unspent in unspents:
            unspent.opt_in_for_RBF()

    if remaining > 0:
        outputs.append((leftover, remaining))

    # Sanity check: If spending from main-/testnet, then all output addresses must also be for main-/testnet.
    for output in outputs:
        dest, amount = output
        vs = get_version(dest)
        if vs and vs != version:
            raise ValueError('Cannot send to ' + vs + 'net address when spending from a ' + version + 'net address.')

    outputs.extend(messages)

    return unspents, outputs


def address_to_scriptpubkey(address):
    # Raise ValueError if we cannot identify the address.
    get_version(address)
    try:
        version = b58decode_check(address)[:1]
    except ValueError:
        witver, data = segwit_decode(address)
        return segwit_scriptpubkey(witver, data)

    if version == MAIN_PUBKEY_HASH or version == TEST_PUBKEY_HASH:
        return OP_DUP + OP_HASH160 + OP_PUSH_20 + address_to_public_key_hash(address) + OP_EQUALVERIFY + OP_CHECKSIG
    elif version == MAIN_SCRIPT_HASH or version == TEST_SCRIPT_HASH:
        return OP_HASH160 + OP_PUSH_20 + address_to_public_key_hash(address) + OP_EQUAL


def construct_outputs(outputs):
    outputs_obj = []

    for data in outputs:
        dest, amount = data

        # P2PKH/P2SH/Bech32
        if amount:
            script_pubkey = address_to_scriptpubkey(dest)

            amount = amount.to_bytes(8, byteorder='little')

        # Blockchain storage
        else:
            script_pubkey = OP_RETURN + len(dest).to_bytes(1, byteorder='little') + dest

            amount = b'\x00\x00\x00\x00\x00\x00\x00\x00'

        outputs_obj.append(TxOut(amount, script_pubkey))

    return outputs_obj


def calculate_preimages(tx_obj, inputs_parameters):
    """Calculates preimages for provided transaction structure and input
    values.

    :param tx_obj: The transaction object used to calculate preimage from using
                   a transaction digest algorithm, such as BIP-143 for Segwit
                   inputs. This transaction object must hence have scriptCodes
                   filled into the corresponding scriptSigs in the inputs.
    :type tx_obj: :object:`~bit.transaction.TxObj`
    :param inputs_parameters: A list of tuples with input index as integer,
                              hash type as integer and a boolean flag to denote
                              if the input is spending from a Segwit output.
                              For example: [(0, 1, True), (2, 1, False), (...)]
    :type inputs_parameters: A `list` of `tuple`
    """

    # Tx object data:
    input_count = int_to_varint(len(tx_obj.TxIn))
    output_count = int_to_varint(len(tx_obj.TxOut))
    output_block = b''.join([bytes(o) for o in tx_obj.TxOut])

    hashPrevouts = double_sha256(b''.join([i.txid + i.txindex for i in tx_obj.TxIn]))
    hashSequence = double_sha256(b''.join([i.sequence for i in tx_obj.TxIn]))
    hashOutputs = double_sha256(output_block)

    preimages = []
    for input_index, hash_type, segwit_input in inputs_parameters:
        # We can only handle hashType == 1:
        if hash_type != HASH_TYPE:
            raise ValueError('Bit only support hashType of value 1.')
        # Calculate prehashes:
        if segwit_input:
            # BIP-143 preimage:
            hashed = sha256(
                tx_obj.version
                + hashPrevouts
                + hashSequence
                + tx_obj.TxIn[input_index].txid
                + tx_obj.TxIn[input_index].txindex
                + tx_obj.TxIn[input_index].script_sig_len
                + tx_obj.TxIn[input_index].script_sig  # scriptCode length
                + tx_obj.TxIn[input_index].sequence  # scriptCode (includes amount)
                + hashOutputs
                + tx_obj.locktime
                + hash_type
            )
        else:
            hashed = sha256(
                tx_obj.version
                + input_count
                + b''.join(ti.txid + ti.txindex + OP_0 + ti.sequence for ti in islice(tx_obj.TxIn, input_index))
                + tx_obj.TxIn[input_index].txid
                + tx_obj.TxIn[input_index].txindex
                + tx_obj.TxIn[input_index].script_sig_len
                + tx_obj.TxIn[input_index].script_sig  # scriptCode length
                + tx_obj.TxIn[input_index].sequence  # scriptCode
                + b''.join(
                    ti.txid + ti.txindex + OP_0 + ti.sequence for ti in islice(tx_obj.TxIn, input_index + 1, None)
                )
                + output_count
                + output_block
                + tx_obj.locktime
                + hash_type
            )
        preimages.append(hashed)
    return preimages


def sign_tx(private_key, tx, *, unspents):
    """Signs inputs in provided transaction object for which unspents
    are provided and can be signed by the private key.

    :param private_key: Private key
    :type private_key: ``PrivateKey`` or ``MultiSig``
    :param tx: Transaction object
    :type tx: ``TxObj``
    :param unspents: For inputs to be signed their corresponding Unspent objects
                     must be provided.
    :type unspents: ``list`` of :class:`~bit.network.meta.Unspent`
    :returns: The signed transaction as hex.
    :rtype: ``str``
    """

    # input_dict contains those unspents that can be signed by private_key,
    # providing additional information for segwit-inputs (the amount to spend)
    input_dict = {}
    try:
        for unspent in unspents:
            if not private_key.can_sign_unspent(unspent):
                continue
            tx_input = hex_to_bytes(unspent.txid)[::-1] + unspent.txindex.to_bytes(4, byteorder='little')
            input_dict[tx_input] = unspent.to_dict()
    except TypeError:
        raise TypeError(
            'Please provide as unspents at least all inputs to be signed with the function call in a list.'
        )

    # Determine input indices to sign from input_dict (allows for transaction batching)
    sign_inputs = [j for j, i in enumerate(tx.TxIn) if i.txid + i.txindex in input_dict]

    segwit_tx = TxObj.is_segwit(tx)
    public_key = private_key.public_key
    public_key_push = script_push(len(public_key))
    hash_type = HASH_TYPE

    # Make input parameters for preimage calculation
    inputs_parameters = []

    # The TxObj in `tx` will below be modified to contain the scriptCodes used
    # for the transaction structure to be signed

    # `input_script_field` copies the scriptSigs for partially signed
    # transactions to later extract signatures from it:
    input_script_field = [tx.TxIn[i].script_sig for i in range(len(tx.TxIn))]

    for i in sign_inputs:
        # Create transaction object for preimage calculation
        tx_input = tx.TxIn[i].txid + tx.TxIn[i].txindex
        segwit_input = input_dict[tx_input]['segwit']
        tx.TxIn[i].segwit_input = segwit_input

        script_code = private_key.scriptcode
        script_code_len = int_to_varint(len(script_code))

        # Use scriptCode for preimage calculation of transaction object:
        tx.TxIn[i].script_sig = script_code
        tx.TxIn[i].script_sig_len = script_code_len

        if segwit_input:
            try:
                tx.TxIn[i].script_sig += input_dict[tx_input]['amount'].to_bytes(8, byteorder='little')

                # For partially signed Segwit transactions the signatures must
                # be extracted from the witnessScript field:
                input_script_field[i] = tx.TxIn[i].witness
            except AttributeError:
                raise ValueError(
                    'Cannot sign a segwit input when the input\'s amount is '
                    'unknown. Maybe no network connection or the input is '
                    'already spent? Then please provide all inputs to sign as '
                    '`Unspent` objects to the function call.'
                )

        inputs_parameters.append([i, hash_type, segwit_input])
    preimages = calculate_preimages(tx, inputs_parameters)

    # Calculate signature scripts:
    for hash, (i, _, segwit_input) in zip(preimages, inputs_parameters):
        signature = private_key.sign(hash) + b'\x01'

        # ------------------------------------------------------------------
        if private_key.instance == 'MultiSig' or private_key.instance == 'MultiSigTestnet':
            # P2(W)SH input

            script_blob = b''
            sigs = {}
            # Initial number of witness items (OP_0 + one signature + redeemscript).
            witness_count = 3
            if input_script_field[i]:
                sig_list = get_signatures_from_script(input_script_field[i])
                # Bitcoin Core convention: Every missing signature is denoted
                # by 0x00. Only used for already partially-signed scriptSigs:
                script_blob += b'\x00' * (private_key.m - len(sig_list) - 1)
                # Total number of witness items when partially or fully signed:
                witness_count = private_key.m + 2
                # For a partially signed input make a dictionary containing
                # all the provided signatures with public-keys as keys:
                for sig in sig_list:
                    for pub in private_key.public_keys:
                        if verify_sig(sig[:-1], hash, pub):
                            # If we already found a valid signature for pubkey
                            # we just overwrite it and don't care.
                            sigs[pub] = sig
                if len(sigs) >= private_key.m:
                    raise ValueError('Transaction is already signed with sufficiently needed signatures.')

            sigs[public_key] = signature

            witness = b''
            # Sort ingthe signatures according to the public-key list:
            for pub in private_key.public_keys:
                if pub in sigs:
                    sig = sigs[pub]
                    length = int_to_varint(len(sig)) if segwit_input else script_push(len(sig))
                    witness += length + sig

            script_sig = b'\x22' + private_key.segwit_scriptcode

            witness = (int_to_varint(witness_count) if segwit_input else b'') + b'\x00' + witness + script_blob
            witness += (
                int_to_varint(len(private_key.redeemscript))
                if segwit_input
                else script_push(len(private_key.redeemscript))
            ) + private_key.redeemscript

            script_sig = script_sig if segwit_input else witness
            witness = witness if segwit_input else b'\x00' if segwit_tx else b''

        # ------------------------------------------------------------------
        else:
            # P2(W)PKH input

            script_sig = b'\x16' + private_key.segwit_scriptcode

            witness = (
                (b'\x02' if segwit_input else b'')
                + len(signature).to_bytes(1, byteorder='little')  # witness counter
                + signature
                + public_key_push
                + public_key
            )

            script_sig = script_sig if segwit_input else witness
            witness = witness if segwit_input else b'\x00' if segwit_tx else b''

        # Providing the signature(s) to the input
        tx.TxIn[i].script_sig = script_sig
        tx.TxIn[i].script_sig_len = int_to_varint(len(script_sig))
        tx.TxIn[i].witness = witness

    return tx.to_hex()


def create_new_transaction(private_key, unspents, outputs):

    version = VERSION_1
    lock_time = LOCK_TIME
    outputs = construct_outputs(outputs)

    # Optimize for speed, not memory, by pre-computing values.
    inputs = []
    for unspent in unspents:
        script_sig = b''  # empty scriptSig for new unsigned transaction.
        txid = hex_to_bytes(unspent.txid)[::-1]
        txindex = unspent.txindex.to_bytes(4, byteorder='little')
        amount = int(unspent.amount).to_bytes(8, byteorder='little')
        sequence = unspent.sequence.to_bytes(4, byteorder='little')
        inputs.append(TxIn(script_sig, txid, txindex, amount=amount, segwit_input=unspent.segwit,
                           sequence=sequence))

    tx_unsigned = TxObj(version, inputs, outputs, lock_time)

    tx = sign_tx(private_key, tx_unsigned, unspents=unspents)
    return tx


"""
Copyright (c) 2016 Ofek Lev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import decimal
from binascii import hexlify
from coincurve.ecdsa import der_to_cdata


class Decimal(decimal.Decimal):
    def __new__(cls, value):
        return super().__new__(cls, str(value))


def chunk_data(data, size):
    return (data[i : i + size] for i in range(0, len(data), size))


def int_to_unknown_bytes(num, byteorder='big'):
    """Converts an int to the least number of bytes as possible."""
    return num.to_bytes((num.bit_length() + 7) // 8 or 1, byteorder)


def bytes_to_hex(bytestr, upper=False):
    hexed = hexlify(bytestr).decode()
    return hexed.upper() if upper else hexed


def hex_to_bytes(hexed):

    if len(hexed) & 1:
        hexed = '0' + hexed

    return bytes.fromhex(hexed)


def int_to_hex(num, upper=False):
    hexed = hex(num)[2:]
    return hexed.upper() if upper else hexed


def hex_to_int(hexed):
    return int(hexed, 16)


def flip_hex_byte_order(string):
    return bytes_to_hex(hex_to_bytes(string)[::-1])


def int_to_varint(val):

    if val < 253:
        return val.to_bytes(1, 'little')
    elif val <= 65535:
        return b'\xfd' + val.to_bytes(2, 'little')
    elif val <= 4294967295:
        return b'\xfe' + val.to_bytes(4, 'little')
    else:
        return b'\xff' + val.to_bytes(8, 'little')


def script_push(val):

    if val <= 75:
        return int_to_unknown_bytes(val)
    elif val < 256:
        return b'\x4c' + int_to_unknown_bytes(val)
    elif val < 65536:
        return b'\x4d' + val.to_bytes(2, 'little')
    else:
        return b'\x4e' + val.to_bytes(4, 'little')


# Slicing functions returning the byte-data-stream splitted
def read_bytes(stream, bytes):
    return stream[0:bytes], stream[bytes:]


def read_var_string(stream):
    size, stream = read_var_int(stream)
    return read_bytes(stream, size)


def read_var_int(stream):
    val = int(bytes_to_hex(stream[0:1]), base=16)
    if val < 253:
        return val, stream[1:]
    return read_as_int(stream[1:], 2 ** (val - 252))


def read_as_int(stream, bytes):
    return int(bytes_to_hex(stream[0:bytes][::-1]), base=16), stream[bytes:]


def read_segwit_string(stream):
    bytes, stream = read_var_int(stream)
    witness, stream = read_bytes(stream, bytes)
    return int_to_varint(bytes) + witness, stream


def get_signatures_from_script(script):
    """Returns a list of signatures retrieved from the provided (partially)
    signed multisig scriptSig.

    :param script: The partially-signed multisig scriptSig.
    :type script: ``bytes``
    :returns: A list of retrieved signature from the provided scriptSig.
    :rtype: A ``list`` of ``bytes`` signatures
    """

    script = script[1:]  # remove the first OP_0
    sigs = []
    while len(script) > 0:
        # Consume script while extracting possible signatures:
        val, script = read_var_int(script)
        potential_sig, script = read_bytes(script, val)
        try:
            # Raises ValueError if argument to `der_to_cdata` is not a
            # DER-encoding of a signature (without the appended SIGHASH).
            der_to_cdata(potential_sig[:-1])
            # We only add if DER-encoded signature:
            sigs.append(potential_sig)
        except ValueError:
            pass
    return sigs


"""
Copyright (c) 2016 Ofek Lev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from coincurve import verify_signature as _vs

from .base58 import b58decode_check, b58encode_check
from .crypto import ripemd160_sha256, sha256
from .curve import x_to_y

from .utils import int_to_unknown_bytes, hex_to_bytes, script_push
from .bech32 import bech32_decode
from .constants import (
    BECH32_MAIN_VERSION_SET,
    BECH32_TEST_VERSION_SET,
    MAIN_PUBKEY_HASH,
    MAIN_SCRIPT_HASH,
    MAIN_PRIVATE_KEY,
    TEST_PUBKEY_HASH,
    TEST_SCRIPT_HASH,
    TEST_PRIVATE_KEY,
    PUBLIC_KEY_UNCOMPRESSED,
    PUBLIC_KEY_COMPRESSED_EVEN_Y,
    PUBLIC_KEY_COMPRESSED_ODD_Y,
    PRIVATE_KEY_COMPRESSED_PUBKEY,
)


def verify_sig(signature, data, public_key):
    """Verifies some data was signed by the owner of a public key.

    :param signature: The signature to verify.
    :type signature: ``bytes``
    :param data: The data that was supposedly signed.
    :type data: ``bytes``
    :param public_key: The public key.
    :type public_key: ``bytes``
    :returns: ``True`` if all checks pass, ``False`` otherwise.
    """
    return _vs(signature, data, public_key)


def address_to_public_key_hash(address):
    # Raise ValueError if we cannot identify the address.
    get_version(address)
    return b58decode_check(address)[1:]


def get_version(address):
    version, _ = bech32_decode(address)
    if version is None:
        version = b58decode_check(address)[:1]
    if version in (MAIN_PUBKEY_HASH, MAIN_SCRIPT_HASH) or version in BECH32_MAIN_VERSION_SET:
        return 'main'
    elif version in (TEST_PUBKEY_HASH, TEST_SCRIPT_HASH) or version in BECH32_TEST_VERSION_SET:
        return 'test'
    else:
        raise ValueError('{} does not correspond to a mainnet nor testnet address.'.format(version))


def bytes_to_wif(private_key, version='main', compressed=False):

    if version == 'test':
        prefix = TEST_PRIVATE_KEY
    else:
        prefix = MAIN_PRIVATE_KEY

    if compressed:
        suffix = PRIVATE_KEY_COMPRESSED_PUBKEY
    else:
        suffix = b''

    private_key = prefix + private_key + suffix

    return b58encode_check(private_key)


def wif_to_bytes(wif):

    private_key = b58decode_check(wif)

    version = private_key[:1]

    if version == MAIN_PRIVATE_KEY:
        version = 'main'
    elif version == TEST_PRIVATE_KEY:
        version = 'test'
    else:
        raise ValueError('{} does not correspond to a mainnet nor testnet address.'.format(version))

    # Remove version byte and, if present, compression flag.
    if len(wif) == 52 and private_key[-1] == 1:
        private_key, compressed = private_key[1:-1], True
    else:
        private_key, compressed = private_key[1:], False

    return private_key, compressed, version


def wif_checksum_check(wif):

    try:
        decoded = b58decode_check(wif)
    except ValueError:
        return False

    if decoded[:1] in (MAIN_PRIVATE_KEY, TEST_PRIVATE_KEY):
        return True

    return False


def public_key_to_address(public_key, version='main'):

    if version == 'test':
        version = TEST_PUBKEY_HASH
    else:
        version = MAIN_PUBKEY_HASH

    length = len(public_key)

    if length not in (33, 65):
        raise ValueError('{} is an invalid length for a public key.'.format(length))

    return b58encode_check(version + ripemd160_sha256(public_key))


def public_key_to_segwit_address(public_key, version='main'):

    if version == 'test':
        version = TEST_SCRIPT_HASH
    else:
        version = MAIN_SCRIPT_HASH

    length = len(public_key)

    if length != 33:
        raise ValueError(
            '{} is an invalid length for a public key. Segwit only uses compressed public keys'.format(length)
        )

    return b58encode_check(version + ripemd160_sha256(b'\x00\x14' + ripemd160_sha256(public_key)))


def multisig_to_redeemscript(public_keys, m):

    if m > len(public_keys):
        raise ValueError('Required signatures cannot be more than the total number of public keys.')

    redeemscript = int_to_unknown_bytes(m + 80)

    for key in public_keys:
        length = len(key)

        if length not in (33, 65):
            raise ValueError('At least one of the provided public keys is of invalid length {}.'.format(length))

        redeemscript += script_push(length) + key

    redeemscript += (
        int_to_unknown_bytes(len(public_keys) + 80) + b'\xae'
    )  # Only works for n = len(public_keys) < 17. OK due to P2SH script-length limitation.

    if len(redeemscript) > 520:
        raise ValueError('The redeemScript exceeds the allowed 520-byte limitation with the number of public keys.')

    return redeemscript


def multisig_to_address(public_keys, m, version='main'):
    if version == 'test':
        version = TEST_SCRIPT_HASH
    else:
        version = MAIN_SCRIPT_HASH

    return b58encode_check(version + ripemd160_sha256(multisig_to_redeemscript(public_keys, m)))


def multisig_to_segwit_address(public_keys, m, version='main'):
    if version == 'test':
        version = TEST_SCRIPT_HASH
    else:
        version = MAIN_SCRIPT_HASH

    return b58encode_check(version + ripemd160_sha256(b'\x00\x20' + sha256(multisig_to_redeemscript(public_keys, m))))


def segwit_scriptpubkey(witver, witprog):
    """Construct a Segwit scriptPubKey for a given witness program."""
    return bytes([witver + 0x50 if witver else 0, len(witprog)] + witprog)


def public_key_to_coords(public_key):

    length = len(public_key)

    if length == 33:
        flag, x = int.from_bytes(public_key[:1], 'big'), int.from_bytes(public_key[1:], 'big')
        y = x_to_y(x, flag & 1)
    elif length == 65:
        x, y = int.from_bytes(public_key[1:33], 'big'), int.from_bytes(public_key[33:], 'big')
    else:
        raise ValueError('{} is an invalid length for a public key.'.format(length))

    return x, y


def coords_to_public_key(x, y, compressed=True):

    if compressed:
        y = PUBLIC_KEY_COMPRESSED_ODD_Y if y & 1 else PUBLIC_KEY_COMPRESSED_EVEN_Y
        return y + x.to_bytes(32, 'big')

    return PUBLIC_KEY_UNCOMPRESSED + x.to_bytes(32, 'big') + y.to_bytes(32, 'big')


def point_to_public_key(point, compressed=True):
    return coords_to_public_key(point.x, point.y, compressed)


"""
Copyright (c) 2016 Ofek Lev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import json
from .utils.crypto import ECPrivateKey, ripemd160_sha256, sha256
from .utils.curve import Point
from .utils.format import (
    bytes_to_wif,
    public_key_to_address,
    public_key_to_coords,
    wif_to_bytes,
    multisig_to_address,
    multisig_to_redeemscript,
    public_key_to_segwit_address,
    multisig_to_segwit_address
)
from .net import Net
from .fees import get_fees
from .tx import (
    calc_txid,
    create_new_transaction,
    sanitize_tx_data,
    sign_tx,
    deserialize,
    address_to_scriptpubkey,
)
from .utils.opcodes import OP_0, OP_PUSH_20, OP_PUSH_32
from .utils.utils import hex_to_bytes, bytes_to_hex, int_to_varint
from .key import Key

class BaseKey:
    """This class represents a point on the elliptic curve secp256k1 and
    provides all necessary cryptographic functionality. You shouldn't use
    this class directly.
    :param wif: A private key serialized to the Wallet Import Format. If the
                argument is not supplied, a new private key will be created.
                The WIF compression flag will be adhered to, but the version
                byte is disregarded. Compression will be used by all new keys.
    :type wif: ``str``
    :raises TypeError: If ``wif`` is not a ``str``.
    """

    def __init__(self, wif=None):
        if wif:
            if isinstance(wif, str):
                private_key_bytes, compressed, version = wif_to_bytes(wif)
                self._pk = ECPrivateKey(private_key_bytes)
            elif isinstance(wif, ECPrivateKey):
                self._pk = wif
                compressed = True
            else:
                raise TypeError('Wallet Import Format must be a string.')
        else:
            self._pk = ECPrivateKey()
            compressed = True

        self._public_point = None
        self._public_key = self._pk.public_key.format(compressed=compressed)

    @property
    def public_key(self):
        """The public point serialized to bytes."""
        return self._public_key

    @property
    def public_point(self):
        """The public point (x, y)."""
        if self._public_point is None:
            self._public_point = Point(*public_key_to_coords(self._public_key))
        return self._public_point

    def sign(self, data):
        """Signs some data which can be verified later by others using
        the public key.
        :param data: The message to sign.
        :type data: ``bytes``
        :returns: A signature compliant with BIP-62.
        :rtype: ``bytes``
        """
        return self._pk.sign(data)

    def verify(self, signature, data):
        """Verifies some data was signed by this private key.
        :param signature: The signature to verify.
        :type signature: ``bytes``
        :param data: The data that was supposedly signed.
        :type data: ``bytes``
        :rtype: ``bool``
        """
        return self._pk.public_key.verify(signature, data)

    def pub_to_hex(self):
        """:rtype: ``str`` """
        return bytes_to_hex(self.public_key)

    def to_hex(self):
        """:rtype: ``str``"""
        return self._pk.to_hex()

    def to_bytes(self):
        """:rtype: ``bytes``"""
        return self._pk.secret

    def to_der(self):
        """:rtype: ``bytes``"""
        return self._pk.to_der()

    def to_pem(self):
        """:rtype: ``bytes``"""
        return self._pk.to_pem()

    def to_int(self):
        """:rtype: ``int``"""
        return self._pk.to_int()

    def is_compressed(self):
        """Returns whether or not this private key corresponds to a compressed
        public key.
        :rtype: ``bool``
        """
        return True if len(self.public_key) == 33 else False

    def __eq__(self, other):
        return self.to_int() == other.to_int()


class PrivateKey(BaseKey):
    """This class represents a Bitcoin private key. ``Key`` is an alias.
    :param wif: A private key serialized to the Wallet Import Format. If the
                argument is not supplied, a new private key will be created.
                The WIF compression flag will be adhered to, but the version
                byte is disregarded. Compression will be used by all new keys.
    :type wif: ``str``
    :raises TypeError: If ``wif`` is not a ``str``.
    """

    def __init__(self, wif=None, network="mainnet"):
        super().__init__(wif=wif)

        if network == "mainnet":
            self.version = 'main'
            self.instance = 'PrivateKey'
        else:
            self.version = 'test'
            self.instance = 'PrivateKeyTestnet'

        self.network = network

        self._address = None
        self._segwit_address = None
        self._scriptcode = None
        self._segwit_scriptcode = None

        self.balance = 0
        self.unspents = []
        self.transactions = []

    @property
    def address(self):
        """The public address you share with others to receive funds."""
        if self._address is None:
            self._address = public_key_to_address(self._public_key, version=self.version)
        return self._address

    @property
    def segwit_address(self):
        """The public segwit nested in P2SH address you share with others to
        receive funds."""
        # Only make segwit address if public key is compressed
        if self._segwit_address is None and self.is_compressed():
            self._segwit_address = public_key_to_segwit_address(self._public_key, version=self.version)
        return self._segwit_address

    @property
    def scriptcode(self):
        self._scriptcode = address_to_scriptpubkey(self.address)
        return self._scriptcode

    @property
    def segwit_scriptcode(self):
        self._segwit_scriptcode = OP_0 + OP_PUSH_20 + ripemd160_sha256(self.public_key)
        return self._segwit_scriptcode

    def can_sign_unspent(self, unspent):
        script = bytes_to_hex(address_to_scriptpubkey(self.address))
        if self.segwit_address:
            segwit_script = bytes_to_hex(address_to_scriptpubkey(self.segwit_address))
            return unspent.script == script or unspent.script == segwit_script
        else:
            return unspent.script == script

    def to_wif(self):
        return bytes_to_wif(self._pk.secret, version=self.version, compressed=self.is_compressed())

    def balance_as(self, currency):
        """Returns your balance as a formatted string in a particular currency.
        :param currency: One of the :ref:`supported currencies`.
        :type currency: ``str``
        :rtype: ``str``
        """
        return self.balance

    def get_balance(self, currency='satoshi'):
        """Fetches the current balance by calling
        :func:`~bit.PrivateKey.get_unspents` and returns it using
        :func:`~bit.PrivateKey.balance_as`.
        :param currency: One of the :ref:`supported currencies`.
        :type currency: ``str``
        :rtype: ``str``
        """
        self.get_unspents()
        return self.balance_as(currency)

    def get_unspents(self):
        """Fetches all available unspent transaction outputs.
        :rtype: ``list`` of :class:`~bit.network.meta.Unspent`
        """
        self.unspents[:] = list(
            map(
                lambda u: u.set_type('p2pkh' if self.is_compressed() else 'p2pkh-uncompressed'),
                Net(network=self.network).utxo(self.address),
            )
        )
        if self.segwit_address:
            self.unspents += list(map(lambda u: u.set_type('np2wkh'), Net(network=self.network).utxo(self.address)))
        self.balance = sum(unspent.amount for unspent in self.unspents)
        return self.unspents

    def get_transactions(self):
        """Fetches transaction history.
        :rtype: ``list`` of ``str`` transaction IDs
        """
        self.transactions[:] = Net(network=self.network).txs(self.address)
        if self.segwit_address:
            self.transactions += Net(network=self.network).txs(self.address)
        return self.transactions

    def create_transaction(
        self,
        outputs,
        fee=None,
        absolute_fee=False,
        leftover=None,
        combine=True,
        message=None,
        unspents=None,
        message_is_hex=False,
        replace_by_fee=False
    ):  # pragma: no cover
        """Creates a signed P2PKH transaction.
        :param outputs: A sequence of outputs you wish to send in the form
                        ``(destination, amount, currency)``. The amount can
                        be either an int, float, or string as long as it is
                        a valid input to ``decimal.Decimal``. The currency
                        must be :ref:`supported <supported currencies>`.
        :type outputs: ``list`` of ``tuple``
        :param fee: The number of satoshi per byte to pay to miners. By default
                    Bit will poll `<https://bitcoinfees.earn.com>`_ and use a fee
                    that will allow your transaction to be confirmed as soon as
                    possible.
        :type fee: ``int``
        :param leftover: The destination that will receive any change from the
                         transaction. By default Bit will send any change to
                         the same address you sent from.
        :type leftover: ``str``
        :param combine: Whether or not Bit should use all available UTXOs to
                        make future transactions smaller and therefore reduce
                        fees. By default Bit will consolidate UTXOs. Note: When
                        setting :param absolute_fee: this is ignored.
        :type combine: ``bool``
        :param message: A message to include in the transaction. This will be
                        stored in the blockchain forever. Due to size limits,
                        each message will be stored in chunks of 40 bytes.
        :type message: ``str``
        :param unspents: The UTXOs to use as the inputs. By default Bit will
                         communicate with the blockchain itself.
        :type unspents: ``list`` of :class:`~bit.network.meta.Unspent`
        :param replace_by_fee: Whether to opt-in for replace-by-fee (BIP 125).
        :type replace_by_fee: ``bool``
        :returns: The signed transaction as hex.
        :rtype: ``str``
        """
        try:
            unspents = unspents or self.get_unspents()
        except ConnectionError:
            raise ConnectionError('All APIs are unreachable. Please provide the unspents to spend from directly.')

        # If at least one input is from segwit the return address is for segwit
        return_address = self.segwit_address if any([u.segwit for u in unspents]) else self.address

        unspents, outputs = sanitize_tx_data(
            unspents,
            outputs,
            (fee or get_fees()) if self.network == "mainnet" else 2,
            leftover or return_address,
            combine=combine,
            message=message,
            absolute_fee=absolute_fee,
            version=self.version,
            message_is_hex=message_is_hex,
            replace_by_fee=replace_by_fee
        )

        return create_new_transaction(self, unspents, outputs)

    def send(
        self,
        outputs,
        fee=None,
        absolute_fee=False,
        leftover=None,
        combine=True,
        message=None,
        unspents=None,
        message_is_hex=False,
        replace_by_fee=False
    ):  # pragma: no cover
        """Creates a signed P2PKH transaction and attempts to broadcast it on
        the blockchain. This accepts the same arguments as
        :func:`~bit.PrivateKey.create_transaction`.
        :param outputs: A sequence of outputs you wish to send in the form
                        ``(destination, amount, currency)``. The amount can
                        be either an int, float, or string as long as it is
                        a valid input to ``decimal.Decimal``. The currency
                        must be :ref:`supported <supported currencies>`.
        :type outputs: ``list`` of ``tuple``
        :param fee: The number of satoshi per byte to pay to miners. By default
                    Bit will poll `<https://bitcoinfees.earn.com>`_ and use a fee
                    that will allow your transaction to be confirmed as soon as
                    possible.
        :type fee: ``int``
        :param leftover: The destination that will receive any change from the
                         transaction. By default Bit will send any change to
                         the same address you sent from.
        :type leftover: ``str``
        :param combine: Whether or not Bit should use all available UTXOs to
                        make future transactions smaller and therefore reduce
                        fees. By default Bit will consolidate UTXOs. Note: When
                        setting :param absolute_fee: this is ignored.
        :type combine: ``bool``
        :param message: A message to include in the transaction. This will be
                        stored in the blockchain forever. Due to size limits,
                        each message will be stored in chunks of 40 bytes.
        :type message: ``str``
        :param unspents: The UTXOs to use as the inputs. By default Bit will
                         communicate with the blockchain itself.
        :type unspents: ``list`` of :class:`~bit.network.meta.Unspent`
        :param replace_by_fee: Whether to opt-in for replace-by-fee (BIP 125).
        :type replace_by_fee: ``bool``
        :returns: The transaction ID.
        :rtype: ``str``
        """

        tx_hex = self.create_transaction(
            outputs,
            fee=(fee or get_fees()) if self.network == "mainnet" else 2,
            absolute_fee=absolute_fee,
            leftover=leftover,
            combine=combine,
            message=message,
            unspents=unspents,
            message_is_hex=message_is_hex,
            replace_by_fee=replace_by_fee
        )

        Net(network=self.network).broadcast(tx_hex)

        return calc_txid(tx_hex)

    @classmethod
    def prepare_transaction(
        cls,
        address,
        outputs,
        compressed=True,
        fee=None,
        absolute_fee=False,
        leftover=None,
        combine=True,
        message=None,
        unspents=None,
        message_is_hex=False,
        replace_by_fee=False
    ):  # pragma: no cover
        """Prepares a P2PKH transaction for offline signing.
        :param address: The address the funds will be sent from.
        :type address: ``str``
        :param outputs: A sequence of outputs you wish to send in the form
                        ``(destination, amount, currency)``. The amount can
                        be either an int, float, or string as long as it is
                        a valid input to ``decimal.Decimal``. The currency
                        must be :ref:`supported <supported currencies>`.
        :type outputs: ``list`` of ``tuple``
        :param compressed: Whether or not the ``address`` corresponds to a
                           compressed public key. This influences the fee.
        :type compressed: ``bool``
        :param fee: The number of satoshi per byte to pay to miners. By default
                    Bit will poll `<https://bitcoinfees.earn.com>`_ and use a fee
                    that will allow your transaction to be confirmed as soon as
                    possible.
        :type fee: ``int``
        :param leftover: The destination that will receive any change from the
                         transaction. By default Bit will send any change to
                         the same address you sent from.
        :type leftover: ``str``
        :param combine: Whether or not Bit should use all available UTXOs to
                        make future transactions smaller and therefore reduce
                        fees. By default Bit will consolidate UTXOs. Note: When
                        setting :param absolute_fee: this is ignored.
        :type combine: ``bool``
        :param message: A message to include in the transaction. This will be
                        stored in the blockchain forever. Due to size limits,
                        each message will be stored in chunks of 40 bytes.
        :type message: ``str``
        :param unspents: The UTXOs to use as the inputs. By default Bit will
                         communicate with the blockchain itself.
        :type unspents: ``list`` of :class:`~bit.network.meta.Unspent`
        :param replace_by_fee: Whether to opt-in for replace-by-fee (BIP 125).
        :type replace_by_fee: ``bool``
        :returns: JSON storing data required to create an offline transaction.
        :rtype: ``str``
        """
        unspents, outputs = sanitize_tx_data(
            unspents or Net(network=self.network).utxo(address),
            outputs,
            (fee or get_fees()) if self.network == "mainnet" else 2,
            leftover or address,
            combine=combine,
            message=message,
            absolute_fee=absolute_fee,
            version='main',
            message_is_hex=message_is_hex,
            replace_by_fee=replace_by_fee
        )

        data = {'unspents': [unspent.to_dict() for unspent in unspents], 'outputs': outputs}

        return json.dumps(data, separators=(',', ':'))

    def sign_transaction(self, tx_data, unspents=None):  # pragma: no cover
        """Creates a signed P2PKH transaction using previously prepared
        transaction data.
        :param tx_data: Hex-encoded transaction or output of :func:`~bit.Key.prepare_transaction`.
        :type tx_data: ``str``
        :param unspents: The UTXOs to use as the inputs. By default Bit will
                         communicate with the blockchain itself.
        :type unspents: ``list`` of :class:`~bit.network.meta.Unspent`
        :returns: The signed transaction as hex.
        :rtype: ``str``
        """
        try:  # Json-tx-data from :func:`~bit.Key.prepare_transaction`
            data = json.loads(tx_data)
            assert unspents is None

            unspents = [Unspent.from_dict(unspent) for unspent in data['unspents']]
            outputs = data['outputs']

            return create_new_transaction(self, unspents, outputs)
        except:  # May be hex-encoded transaction using batching:
            try:
                unspents = unspents or self.get_unspents()
            except ConnectionError:
                raise ConnectionError(
                    'All APIs are unreachable. Please provide the unspent '
                    'inputs as unspents directly to sign this transaction.'
                )

            tx_data = deserialize(tx_data)
            return sign_tx(self, tx_data, unspents=unspents)

    @classmethod
    def from_hex(cls, hexed):
        """
        :param hexed: A private key previously encoded as hex.
        :type hexed: ``str``
        :rtype: :class:`~bit.PrivateKey`
        """
        return PrivateKey(ECPrivateKey.from_hex(hexed))

    @classmethod
    def from_bytes(cls, bytestr):
        """
        :param bytestr: A private key previously encoded as hex.
        :type bytestr: ``bytes``
        :rtype: :class:`~bit.PrivateKey`
        """
        return PrivateKey(ECPrivateKey(bytestr))

    @classmethod
    def from_der(cls, der):
        """
        :param der: A private key previously encoded as DER.
        :type der: ``bytes``
        :rtype: :class:`~bit.PrivateKey`
        """
        return PrivateKey(ECPrivateKey.from_der(der))

    @classmethod
    def from_pem(cls, pem):
        """
        :param pem: A private key previously encoded as PEM.
        :type pem: ``bytes``
        :rtype: :class:`~bit.PrivateKey`
        """
        return PrivateKey(ECPrivateKey.from_pem(pem))

    @classmethod
    def from_int(cls, num):
        """
        :param num: A private key in raw integer form.
        :type num: ``int``
        :rtype: :class:`~bit.PrivateKey`
        """
        return PrivateKey(ECPrivateKey.from_int(num))

    def __repr__(self):
        return '<PrivateKey: {}>'.format(self.address)


from .external import Blockchair, Blockstream

class Net:
    def __init__(self, provider='blockstream', network='mainnet'):
        provider = provider.lower()
        if provider == 'blockchair':
            self.provider = Blockchair(network=network)
        elif provider == "blockstream":
            self.provider = Blockstream(network=network)
        else:
            raise ValueError("Invalid provider")

    def balance(self, address, unit='satoshi'):
        r = self.provider.balance(address, unit)
        return r

    def txs(self, address):
        r = self.provider.txs(address)
        return r

    def utxo(self, address):
        r = self.provider.utxo(address)
        return r

    def broadcast(self, tx_hex):
        r = self.provider.broadcast(tx_hex)
        return r


from oogway.operation import operation
import decimal

def test_add_sats():
    op_1 = operation.add_sats(1, 1, 1, unit="satoshi")
    op_2 = operation.add_sats(50000000, 50000000, unit="btc")

    assert op_1 == 3
    assert op_2 == decimal.Decimal('1.00000000')
    assert type(op_1) == int
    assert type(op_2) == decimal.Decimal

def test_substract_sats():
    op_1 = operation.substract_sats(10, 1, unit="satoshi")
    op_2 = operation.substract_sats(200000000, 100000000, unit="btc")

    assert op_1 == 9
    assert op_2 == decimal.Decimal('1.00000000')
    assert type(op_1) == int
    assert type(op_2) == decimal.Decimal

def test_add_btc():
    op_1 = operation.add_btc(1, 1, unit="btc")
    op_2 = operation.add_btc(1, 1, unit="satoshi")

    assert op_1 == decimal.Decimal('2.00000000')
    assert op_2 == 200000000
    assert type(op_1) == decimal.Decimal
    assert type(op_2) == int

def test_substract_btc():
    op_1 = operation.substract_btc(2, 1, unit="btc")
    op_2 = operation.substract_btc(2, 1, unit="satoshi")

    assert op_1 == decimal.Decimal('1.00000000')
    assert op_2 == 100000000
    assert type(op_1) == decimal.Decimal
    assert type(op_2) == int



import binascii
import hashlib
import ecdsa
from .utils.base58 import b58encode
from .utils import bech32
from .utils.crypto import doublehash256

def hash160(v):
    r = hashlib.new('ripemd160')
    r.update(hashlib.sha256(v).digest())
    return r

def SECP256k1(digest):
    sk = ecdsa.SigningKey.from_string(digest, curve=ecdsa.SECP256k1)
    return sk.get_verifying_key()

def gen_p2pkh(_hash, testnet=False):
    """generate p2pkh address (prefix is 1)"""
    if testnet == False:
        prefix_a = b'\x04'
        prefix_b = b'\x00'
    else:
        prefix_a = b'\x04'
        prefix_b = b'\x6F'
    address = gen_unc_pub(_hash, prefix_a, prefix_b)
    return address

def gen_p2sh(_hash, testnet=False):
    """generate p2sh address (prefix is 3)"""
    if testnet == False:
        prefix = b'\x05'
    else:
        prefix = b'\xC4'
    address = _gen_p2sh(_hash, prefix)
    return address

def _gen_p2sh(_hash, prefix):
    """generate p2sh address (prefix is 3)"""
    prefix_redeem = b'\x00\x14'
    p = gen_c_pub(_hash, b'\x02', b'\x03')
    redeem_script = hash160(prefix_redeem + hash160(p).digest()).digest()
    m = prefix + redeem_script
    checksum = doublehash256(m).digest()[:4]
    address = b58encode(m + checksum)
    return address

def gen_bech32(_hash, testnet=False):
    """generate bech32 P2WSH address (prefix is bc1 | tb1)"""
    if testnet == False:
        prefix_a = b'\x02'
        prefix_b = b'\x03'
        address_prefix = 'bc'
    else:
        prefix_a = b'\x02'
        prefix_b = b'\x14'
        address_prefix = 'tb'
    p = gen_c_pub(_hash, prefix_a, prefix_b)
    redeem_script_P2WSH = hashlib.sha256(p).digest()
    address = str(bech32.encode(address_prefix, 0x00, redeem_script_P2WSH))
    return address

def gen_unc_pub(_hash, prefix_a, prefix_b):
    """generate uncompressed pubkey from hash"""
    digest = _hash.digest()
    p = prefix_a + SECP256k1(digest).to_string()
    pubkey = str(binascii.hexlify(p).decode('utf-8'))
    _hash160 = hash160(p)
    m = prefix_b + _hash160.digest()
    checksum = doublehash256(m).digest()[:4]

    unc = b58encode(m + checksum)
    return unc

def _gen_unc_pub(_hash):
    digest = _hash.digest()
    p = b'\x04' + SECP256k1(digest).to_string()
    pubkey = str(binascii.hexlify(p).decode('utf-8'))
    return pubkey

def gen_c_pub(_hash, prefix_even, prefix_odd):
    """generate compressed pubkey from hash"""
    prefix_a = prefix_odd
    digest = _hash.digest()
    ecdsa_digest = SECP256k1(digest).to_string()
    x_coord = ecdsa_digest[:int(len(ecdsa_digest) / 2)]
    y_coord = ecdsa_digest[int(len(ecdsa_digest) / 2):]
    if (int(binascii.hexlify(y_coord), 16) % 2 == 0): prefix_a = prefix_even
    p = prefix_a + x_coord
    c = str(binascii.hexlify(p).decode('utf-8'))
    return p

def _gen_c_pub(_hash):
    prefix_a = b'\x03'
    prefix_even = b'\x02' 
    digest = _hash.digest()
    ecdsa_digest = SECP256k1(digest).to_string()
    x_coord = ecdsa_digest[:int(len(ecdsa_digest) / 2)]
    y_coord = ecdsa_digest[int(len(ecdsa_digest) / 2):]
    if (int(binascii.hexlify(y_coord), 16) % 2 == 0): prefix_a = prefix_even
    p = prefix_a + x_coord
    c = str(binascii.hexlify(p).decode('utf-8'))
    return c

class Address:
    def __init__(self, _hash, testnet=False):
        self.pubkey = _gen_unc_pub(_hash)
        self.pubkey_c = _gen_c_pub(_hash)
        self.p2pkh = gen_p2pkh(_hash, testnet)
        self.p2sh = gen_p2sh(_hash, testnet)
        self.bech32 = gen_bech32(_hash, testnet)



from oogway.validate import _bech32check, p2pkh_val, p2sh_val, bech32_val, validate
from tests.data import *

def test__bech32check():
    """validate Bech32 format by encoding only"""
    assert _bech32check(BECH32_ADDRESS) == True
    assert _bech32check(INVALID_BECH32_ADDRESS) == False

def test_p2pkh_val():
    """validate P2PKH format"""
    assert p2pkh_val(P2PKH_ADDRESS) == True
    assert p2sh_val(INVALID_P2PKH_ADDRESS) == False

def test_p2sh_val():
    """validate P2SH format"""
    assert p2sh_val(P2SH_ADDRESS) == True
    assert p2sh_val(INVALID_P2SH_ADDRESS) == False

def test_bech32_val():
    """validate Bech32 format"""
    assert bech32_val(BECH32_ADDRESS) == True
    assert bech32_val(INVALID_BECH32_ADDRESS) == False

def test_address_format():
    """returns address format"""
    bech32 = validate.address_format(BECH32_ADDRESS)
    p2sh = validate.address_format(P2SH_ADDRESS)
    p2pkh = validate.address_format(P2PKH_ADDRESS)

    assert bech32 == 'Bech32'
    assert p2sh == 'P2SH'
    assert p2pkh == 'P2PKH'



import click
from oogway.net import Net
from oogway.convert import convert

@click.command()
@click.argument('addresses', nargs=-1)
@click.option('--network', default="mainnet", help="Network (mainnet | testnet)")
def balance(addresses, network):
    """Get addresses balance"""
    n = Net(provider="blockstream", network=network)
    
    for address in addresses:
        bal_sat = n.balance(address)
        bal_btc = convert.to_btc(bal_sat, string=True)

        click.echo("\n[%s] %s Satoshis (%s BTC)\n" % (address, bal_sat, bal_btc))



from collections import deque
from hashlib import sha256
from .crypto import double_sha256_checksum
from .utils import int_to_unknown_bytes

BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
BASE58_ALPHABET_LIST = list(BASE58_ALPHABET)
BASE58_ALPHABET_INDEX = {char: index for index, char in enumerate(BASE58_ALPHABET)}

def b58encode(bytestr):
    """encode to base58"""
    alphabet = BASE58_ALPHABET_LIST
    encoded = deque()
    append = encoded.appendleft
    _divmod = divmod

    num = int.from_bytes(bytestr, 'big')

    while num > 0:
        num, rem = _divmod(num, 58)
        append(alphabet[rem])
    
    encoded = ''.join(encoded)
    pad = 0
    for byte in bytestr:
        if byte == 0:
            pad += 1
        else:
            break
    
    return '1' * pad + encoded

def b58encode_check(bytestr):
    return b58encode(bytestr + double_sha256_checksum(bytestr))

def b58decode(bytestr, length):
    """decode from base58"""
    n = 0
    for char in bytestr:
        n = n * 58 + BASE58_ALPHABET.index(char)
    return n.to_bytes(length, 'big')

def b58check(bytestr):
    """check if base58 encoded address is valid"""
    try:
        bcbytes = b58decode(bytestr, 25)
        return bcbytes[-4:] == sha256(sha256(bcbytes[:-4]).digest()).digest()[:4]
    except Exception:
        return False

def _b58decode(string):

    alphabet_index = BASE58_ALPHABET_INDEX

    num = 0

    try:
        for char in string:
            num *= 58
            num += alphabet_index[char]
    except KeyError:
        raise ValueError('"{}" is an invalid base58 encoded character.'.format(char)) from None

    bytestr = int_to_unknown_bytes(num)

    pad = 0
    for char in string:
        if char == '1':
            pad += 1
        else:
            break

    return b'\x00' * pad + bytestr

def b58decode_check(string):

    decoded = _b58decode(string)
    shortened = decoded[:-4]
    decoded_checksum = decoded[-4:]
    hash_checksum = double_sha256_checksum(shortened)

    if decoded_checksum != hash_checksum:
        raise ValueError(
            'Decoded checksum {} derived from "{}" is not equal to hash '
            'checksum {}.'.format(decoded_checksum, string, hash_checksum)
        )

    return shortened


import hashlib
from .seed.gen import Mnemonic
from .utils.crypto import doublehash256
from .utils.base58 import b58encode
from .address import Address

def gen_hex(_hash):
    """converts private key to hex format"""
    _hex = _hash.hexdigest()
    return _hex
    
def gen_wif(_hashdigest, testnet=False):
    """converts private key to wif and compressed wif format"""
    if testnet == False:
        prefix = b'\x80'
        suffix = b'\x01'
    else:
        prefix = b'\xEF'
        suffix = b'\x01'

    d = prefix + _hashdigest
    checksum = doublehash256(d).digest()[:4]
    checksum_c = doublehash256(d + suffix).digest()[:4]

    wif = b58encode(d + checksum)
    wif_c = b58encode(d + suffix + checksum_c)

    return wif, wif_c

class Key:
    def __init__(self, mnemonic=None, testnet=False, mnemonic_strength=256, passphrase=""):
        self.hex = ""
        self.wif = ""
        self.wif_c = ""
        self.hash = ""
        self.hashdigest = ""
        self.seed = None
        self.mnemonic = mnemonic
        self.testnet = testnet
        
        mn = Mnemonic()
        if self.mnemonic == None:
            self.mnemonic = mn.generate(strength=mnemonic_strength)
        else:
            pass
        self.seed = mn.to_seed(mnemonic=self.mnemonic, passphrase=passphrase)

        self.hash = hashlib.sha256(self.seed)
        self.hashdigest = self.hash.digest()

        # format private keys
        self.hex = gen_hex(self.hash)
        self.wif, self.wif_c = gen_wif(self.hashdigest, testnet=testnet)
        
    def address(self, _format='Bech32'):
        """returns generated address in selected format | default is bech32"""
        _address = Address(self.hash, testnet=self.testnet)
        _format = _format.lower()
        if _format == 'bech32':
            a = _address.bech32
        elif _format == 'p2pkh':
            a = _address.p2pkh
        elif _format == 'p2sh':
            a = _address.p2sh
        else:
            raise ValueError("Address format %s does not exist" % _format)
        
        return a

    def pubkey(self, _format='unc'):
        """returns generated public key in selected format | default is uncompressed"""
        _pubkey = Address(self.hash, testnet=self.testnet)
        _format = _format.lower()
        if _format == 'unc':
            pb = _pubkey.pubkey
        elif _format == 'c':
            pb = _pubkey.pubkey_c
        else:
            raise ValueError("Public key format %s does not exist" % _format)

        return pb



import os
import hashlib
import binascii
import hmac
from .base58 import b58encode

class Mnemonic:
    def __init__(self):
        with open("%s/english.txt" % (self._get_directory()), "r", encoding="utf-8") as f:
            self.wordlist = [w.strip() for w in f.readlines()]
    
    @classmethod
    def _get_directory(cls):
        return os.path.join(os.path.dirname(__file__), "wordlist")

    def generate(self, strength=128):
        if strength not in [128, 160, 192, 224, 256]:
            raise ValueError("Strength should be one of the following [128, 160, 192, 224, 256]")
        return self.to_mnemonic(os.urandom(strength // 8))

    def to_mnemonic(self, data):
        h = hashlib.sha256(data).hexdigest()
        b = (
        bin(int(binascii.hexlify(data), 16))[2:].zfill(len(data) * 8)
        + bin(int(h, 16))[2:].zfill(256)[: len(data) * 8 // 32]
        )
        result = []
        for i in range(len(b) // 11):
            idx = int(b[i * 11 : (i + 1) * 11], 2)
            result.append(self.wordlist[idx])
        result_phrase = " ".join(result)
        return result_phrase
    
    @classmethod
    def to_seed(cls, mnemonic, passphrase=""):
        passphrase = "mnemonic"+passphrase
        mnemonic = mnemonic.encode("utf-8")
        passphrase = passphrase.encode("utf-8")
        stretched = hashlib.pbkdf2_hmac("sha512", mnemonic, passphrase, 2048)
        return stretched[:64]

    @classmethod
    def to_hd_master_key(cls, seed):
        if len(seed) != 64:
            raise ValueError("Provided seed should have length of 64")
        seed = hmac.new(b"Bitcoin seed", seed, digestmod=hashlib.sha512).digest()
        xprv = b"\x04\x88\xad\xe4"
        xprv += b"\x00" * 9
        xprv += seed[32:]
        xprv += b"\x00" + seed[:32]

        hashed_xprv = hashlib.sha256(xprv).digest()
        hashed_xprv = hashlib.sha256(hashed_xprv).digest()

        xprv += hashed_xprv[:4]

        return b58encode(xprv)



def test_key_module():
    try:
        from oogway import Key
        imported = True
    except ImportError:
        imported = False

    assert imported == True

def test_validate_module():
    try:
        from oogway import validate
        imported = True
    except ImportError:
        imported = False
    
    assert imported == True

def test_convert_module():
    try:
        from oogway import convert
        imported = True
    except ImportError:
        imported = False
    
    assert imported == True

def test_operation_module():
    try:
        from oogway import operation
        imported = True
    except ImportError:
        imported = False

    assert imported == True

def test_fees_module():
    try:
        from oogway import get_fees
        imported = True
    except ImportError:
        imported = False

    assert imported == True

def test_request_module():
    try:
        from oogway import request_payment, parse_request
        imported = True
    except ImportError:
        imported = False

    assert imported == True



VERSION_1 = 0x01 .to_bytes(4, byteorder='little')
VERSION_2 = 0x02 .to_bytes(4, byteorder='little')
MARKER = b'\x00'
FLAG = b'\x01'
SEQUENCE = 0xFFFFFFFF .to_bytes(4, byteorder='little')
LOCK_TIME = 0x00 .to_bytes(4, byteorder='little')
HASH_TYPE = 0x01 .to_bytes(4, byteorder='little')

MESSAGE_LIMIT = 80

# Address formats:
BECH32_VERSION_SET = ('bc', 'tb', 'bcrt')
BECH32_MAIN_VERSION_SET = BECH32_VERSION_SET[:1]
BECH32_TEST_VERSION_SET = BECH32_VERSION_SET[1:]
MAIN_PUBKEY_HASH = b'\x00'
MAIN_SCRIPT_HASH = b'\x05'
TEST_PUBKEY_HASH = b'\x6f'
TEST_SCRIPT_HASH = b'\xc4'

# Keys:
MAIN_PRIVATE_KEY = b'\x80'
MAIN_BIP32_PUBKEY = b'\x04\x88\xb2\x1e'
MAIN_BIP32_PRIVKEY = b'\x04\x88\xad\xe4'
TEST_PRIVATE_KEY = b'\xef'
TEST_BIP32_PUBKEY = b'\x045\x87\xcf'
TEST_BIP32_PRIVKEY = b'\x045\x83\x94'
PUBLIC_KEY_UNCOMPRESSED = b'\x04'
PUBLIC_KEY_COMPRESSED_EVEN_Y = b'\x02'
PUBLIC_KEY_COMPRESSED_ODD_Y = b'\x03'
PRIVATE_KEY_COMPRESSED_PUBKEY = b'\x01'


import click
from oogway.convert import convert

@click.command()
@click.argument('btc', nargs=1)
def convert_to_satoshi(btc):
    """convert from BTC to satoshi"""
    conv = convert.to_satoshi(btc, string=True)
    click.echo('\n%s BTC is %s Satoshi\n' % (btc, conv))

@click.command()
@click.argument('satoshi', nargs=1)
def convert_to_btc(satoshi):
    """convert from satoshi to BTC"""
    conv = convert.to_btc(satoshi, string=True)
    click.echo('\n%s Satoshi is %s BTC\n' % (satoshi, conv))



# valid addresses
P2PKH_ADDRESS = "1H36XJJRdzbTJ8yqcR6HBvSTuJJ5dvf5wC"
P2SH_ADDRESS = "3CKnc9a1da7Nnh3SEzvVZe7LfkBq2pYTM2"
BECH32_ADDRESS = "bc1qpfa829hg70yhjw7sg8t95vjtz7gg38k3kfm3ny"

# valid testnet addresses
P2PKH_ADDRESS_TESTNET = "mwZ3pMPQT22i5FTTKz4f1qenmHtncVQqCV"
P2SH_ADDRESS_TESTNET = "2N3szftW3F2cizUfyv8YNBb6bt6PzsUTi92"
BECH32_ADDRESS_TESTNET = "tb1qdh780ud5vtn00327z7qzhjd9nv5fm8qu6qkg2d"

# invalid addresses
INVALID_P2PKH_ADDRESS = "1BvBMSEYstWetqTFn5Au4m4GF07xJaNVN2"
INVALID_P2SH_ADDRESS = "3CKnc9a1da7Nnh3SEzvVZe7LfkBq2IYTM2"
INVALID_BECH32_ADDRESS = "bc1qpfa829hg70yhjw7sg8t95vjtz7gg38k3kf03ny"

# mnemonic phrases
MNEMONIC_1 = "behind view duty stairs globe provide tired poem balance advance early husband unhappy final crunch parade cruise jacket trap bread mention cannon decide alpha"
MNEMONIC_2 = "solid favorite loop sustain grit hub attitude empower prize push extend humor"
MNEMONIC_3 = "chalk range lounge edit arrive erupt dirt voyage buffalo practice find ecology"



import click
from oogway.net import Net
from oogway.tx import calc_txid

@click.command()
@click.option('--network', default='mainnet', help="Broadcasts transaction on selected network (mainnet | testnet)")
@click.argument('tx_hex', nargs=1)
def broadcast(network, tx_hex):
    n = Net(network=network)
    tx_hash = n.broadcast(tx_hex)

    if tx_hash == False:
        click.echo('\nTransaction failed to broadcast.\n')
    elif tx_hash == True:
        txid = calc_txid(tx_hex)
        click.echo('\nTransaction broadcasted: %s\n' % txid)



import hashlib
from binascii import hexlify
from hashlib import new, sha256 as _sha256
from coincurve import PrivateKey as ECPrivateKey, PublicKey as ECPublicKey

def doublehash256(v):
    return hashlib.sha256(hashlib.sha256(v).digest())

def bytes_to_hex(bytestr, upper=False):
    hexed = hexlify(bytestr).decode()
    return hexed.upper() if upper else hexed



def sha256(bytestr):
    return _sha256(bytestr).digest()


def double_sha256(bytestr):
    return _sha256(_sha256(bytestr).digest()).digest()


def double_sha256_checksum(bytestr):
    return double_sha256(bytestr)[:4]


def ripemd160_sha256(bytestr):
    return new('ripemd160', sha256(bytestr)).digest()


hash160 = ripemd160_sha256


import decimal

ONE_BTC_SAT = int(1e8)

class convert:

    @staticmethod
    def to_satoshi(amount, string=False):
        """convert from btc to satoshi"""
        if type(amount) == str:
            amount = float(amount)
        if amount < 0:
            raise ArithmeticError("BTC amount must be a positive number")

        safe = round(amount, 8)
        safe = "{:.8f}".format(amount)
        safe = float(safe) * ONE_BTC_SAT
        safe = int(safe)
        
        if string == False:
            pass
        else:
            safe = str(safe)

        return safe

    @staticmethod
    def to_btc(amount, string=False):
        """convert from satoshi to btc"""
        if type(amount) == str:
            amount = int(amount)
        d = decimal.Decimal(str(amount))
        precision = d.as_tuple().exponent
        
        if precision != 0:
            raise ArithmeticError("Satoshi amount must be an integer")
        if amount < 0:
            raise ArithmeticError("Satoshi amount must be a positive number")

        safe = amount / ONE_BTC_SAT
        safe = round(safe, 8)
        if string == False:
            safe = decimal.Decimal(safe)
            safe = round(safe, 8)
        else:
            safe = "{:.8f}".format(safe)
        
        return safe



import pytest
from oogway.key import Key
from tests.data import *
from oogway.validate import validate
import _hashlib

keys = []
def gen_keys(mnemonic=None, _range=10, _append=True):
    i_keys = []
    for _ in range(_range):
        key = Key(mnemonic)
        if _append == True:
            keys.append(key)
        else:
            i_keys.append(key)
    return i_keys

gen_keys()

@pytest.mark.parametrize("key", keys)
def test_key_generation(key):
    """check key formats type"""
    assert type(key.hash) is _hashlib.HASH
    assert type(key.wif) is str
    assert type(key.hex) is str
    assert type(key.seed) is bytes
    assert type(key.hashdigest) is bytes
    assert type(key.mnemonic) is str

@pytest.mark.parametrize("key", keys)
def test_address_generation(key):
    """validate addresses format"""
    bech32 = key.address('bech32')
    p2pkh = key.address('p2pkh')
    p2sh = key.address('p2sh')
    default_address = key.address()
    
    assert validate.is_bech32(bech32) == True
    assert validate.is_p2pkh(p2pkh) == True
    assert validate.is_p2sh(p2sh) == True
    assert validate.is_bech32(default_address) == True

def test_generation_regularity():
    """check if mnemonic product is regular"""
    _keys = gen_keys(MNEMONIC_1, 3, False)
    wifs = []
    bech32_addresses = []
    p2pkh_addresses = []
    for key in _keys:
        wifs.append(key.wif)
        bech32_addresses.append(key.address('bech32'))
        p2pkh_addresses.append(key.address('p2pkh'))

    wif_result = wifs.count(wifs[0]) == len(wifs)
    bech32_address_result = bech32_addresses.count(bech32_addresses[0]) == len(bech32_addresses)
    p2pkh_address_result = p2pkh_addresses.count(p2pkh_addresses[0]) == len(p2pkh_addresses)

    assert wif_result == True
    assert bech32_address_result == True
    assert p2pkh_address_result == True

@pytest.mark.parametrize("key", keys)
def test_pubkey_format(key):
    pubkey = key.pubkey('unc')
    pubkey_len = len(pubkey)
    
    pubkey_c = key.pubkey('c')
    pubkey_c_len = len(pubkey_c)

    assert pubkey_len == 130
    assert pubkey_c_len == 66
    assert pubkey[0] == '0'
    assert pubkey_c[0] == '0'


class script(object):
    OP_0 = 0x00
    OP_FALSE = OP_0
    OP_PUSHDATA1 = 0x4c
    OP_PUSHDATA2 = 0x4d
    OP_PUSHDATA4 = 0x4e
    OP_1NEGATE = 0x4f
    OP_RESERVED = 0x50
    OP_1 = 0x51
    OP_TRUE=OP_1
    OP_2 = 0x52
    OP_3 = 0x53
    OP_4 = 0x54
    OP_5 = 0x55
    OP_6 = 0x56
    OP_7 = 0x57
    OP_8 = 0x58
    OP_9 = 0x59
    OP_10 = 0x5a
    OP_11 = 0x5b
    OP_12 = 0x5c
    OP_13 = 0x5d
    OP_14 = 0x5e
    OP_15 = 0x5f
    OP_16 = 0x60

    # control
    OP_NOP = 0x61
    OP_VER = 0x62
    OP_IF = 0x63
    OP_NOTIF = 0x64
    OP_VERIF = 0x65
    OP_VERNOTIF = 0x66
    OP_ELSE = 0x67
    OP_ENDIF = 0x68
    OP_VERIFY = 0x69
    OP_RETURN = 0x6a

    # stack ops
    OP_TOALTSTACK = 0x6b
    OP_FROMALTSTACK = 0x6c
    OP_2DROP = 0x6d
    OP_2DUP = 0x6e
    OP_3DUP = 0x6f
    OP_2OVER = 0x70
    OP_2ROT = 0x71
    OP_2SWAP = 0x72
    OP_IFDUP = 0x73
    OP_DEPTH = 0x74
    OP_DROP = 0x75
    OP_DUP = 0x76
    OP_NIP = 0x77
    OP_OVER = 0x78
    OP_PICK = 0x79
    OP_ROLL = 0x7a
    OP_ROT = 0x7b
    OP_SWAP = 0x7c
    OP_TUCK = 0x7d

    # splice ops
    OP_CAT = 0x7e
    OP_SUBSTR = 0x7f
    OP_LEFT = 0x80
    OP_RIGHT = 0x81
    OP_SIZE = 0x82

    # bit logic
    OP_INVERT = 0x83
    OP_AND = 0x84
    OP_OR = 0x85
    OP_XOR = 0x86
    OP_EQUAL = 0x87
    OP_EQUALVERIFY = 0x88
    OP_RESERVED1 = 0x89
    OP_RESERVED2 = 0x8a

    # numeric
    OP_1ADD = 0x8b
    OP_1SUB = 0x8c
    OP_2MUL = 0x8d
    OP_2DIV = 0x8e
    OP_NEGATE = 0x8f
    OP_ABS = 0x90
    OP_NOT = 0x91
    OP_0NOTEQUAL = 0x92

    OP_ADD = 0x93
    OP_SUB = 0x94
    OP_MUL = 0x95
    OP_DIV = 0x96
    OP_MOD = 0x97
    OP_LSHIFT = 0x98
    OP_RSHIFT = 0x99

    OP_BOOLAND = 0x9a
    OP_BOOLOR = 0x9b
    OP_NUMEQUAL = 0x9c
    OP_NUMEQUALVERIFY = 0x9d
    OP_NUMNOTEQUAL = 0x9e
    OP_LESSTHAN = 0x9f
    OP_GREATERTHAN = 0xa0
    OP_LESSTHANOREQUAL = 0xa1
    OP_GREATERTHANOREQUAL = 0xa2
    OP_MIN = 0xa3
    OP_MAX = 0xa4

    OP_WITHIN = 0xa5

    # crypto
    OP_RIPEMD160 = 0xa6
    OP_SHA1 = 0xa7
    OP_SHA256 = 0xa8
    OP_HASH160 = 0xa9
    OP_HASH256 = 0xaa
    OP_CODESEPARATOR = 0xab
    OP_CHECKSIG = 0xac
    OP_CHECKSIGVERIFY = 0xad
    OP_CHECKMULTISIG = 0xae
    OP_CHECKMULTISIGVERIFY = 0xaf

    # expansion
    OP_NOP1 = 0xb0
    OP_CHECKLOCKTIMEVERIFY = 0xb1
    OP_NOP2 = OP_CHECKLOCKTIMEVERIFY
    OP_CHECKSEQUENCEVERIFY = 0xb2
    OP_NOP3 = OP_CHECKSEQUENCEVERIFY
    OP_NOP4 = 0xb3
    OP_NOP5 = 0xb4
    OP_NOP6 = 0xb5
    OP_NOP7 = 0xb6
    OP_NOP8 = 0xb7
    OP_NOP9 = 0xb8
    OP_NOP10 = 0xb9

    OP_INVALIDOPCODE = 0xff

# lib
OP_0 = b'\x00'
OP_CHECKLOCKTIMEVERIFY = b'\xb1'
OP_CHECKSIG = b'\xac'
OP_CHECKMULTISIG = b'\xae'
OP_DUP = b'v'
OP_EQUALVERIFY = b'\x88'
OP_HASH160 = b'\xa9'
OP_PUSH_20 = b'\x14'
OP_PUSH_32 = b'\x20'
OP_RETURN = b'\x6a'
OP_EQUAL = b'\x87'


import requests
from json.decoder import JSONDecodeError

fees_endpoint_blockstream = "https://blockstream.info/api/fee-estimates"
fees_endpoint_earn = "https://bitcoinfees.earn.com/api/v1/fees/recommended"
HARDCODED_FAST = 100
HARDCODED_THREE_BLOCKS = 95
HARDCODED_SIX_BLOCKS = 55

def get_fees(timeframe="fastest", provider="blockstream"):
    """get fees in satoshis per byte"""
    timeframe = timeframe.lower()
    if timeframe == "fastest":
        tf_bs = "1"
        tf_earn = "fastestFee"
        default_fee = HARDCODED_FAST
    elif timeframe == "3":
        tf_bs = "3"
        tf_earn = "halfHourFee"
        default_fee = HARDCODED_THREE_BLOCKS
    elif timeframe == "6":
        tf_bs = "6"
        tf_earn = "hourFee"
        default_fee = HARDCODED_SIX_BLOCKS
    else:
        raise ValueError("The specified timeframe %s does not exist" % timeframe)
    
    provider = provider.lower()
    if provider == "blockstream":
        endpoint = fees_endpoint_blockstream
    elif provider == "earn":
        endpoint = fees_endpoint_earn
    else:
        raise ValueError("The specified provider %s does not exist" % provider)

    fee = requests.get(endpoint)
    fee = fee.json()
    try:
        if provider == "blockstream":
            fee = int(fee[tf_bs])
        elif provider == "earn":
            fee = fee[tf_earn]
            
    except (KeyError, JSONDecodeError):
        fee = default_fee
    return fee



"""
Copyright (c) 2016 Ofek Lev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .constants import SEQUENCE

TX_TRUST_LOW = 1
TX_TRUST_MEDIUM = 6
TX_TRUST_HIGH = 30

UNSPENT_TYPES = {
    # Dictionary containing as keys known unspent types and as value a
    # dictionary containing information if spending uses a witness
    # program (Segwit) and its estimated scriptSig size.
    'unknown': {'segwit': None, 'vsize': 180},  # Unknown type
    'p2pkh-uncompressed': {'segwit': False, 'vsize': 180},  # Legacy P2PKH using  # uncompressed keys
    'p2pkh': {'segwit': False, 'vsize': 148},  # Legacy P2PKH
    'p2sh': {'segwit': False, 'vsize': 298},  # Legacy P2SH (vsize corresponds to a 2-of-3 multisig input)
    'np2wkh': {'segwit': True, 'vsize': 91},  # (Nested) P2SH-P2WKH
    'np2wsh': {'segwit': True, 'vsize': 140},  # (Nested) P2SH-P2WSH (vsize corresponds to a 2-of-3 multisig input)
    'p2wkh': {'segwit': True, 'vsize': 68},  # Bech32 P2WKH -- Not yet supported to sign
    'p2wsh': {
        'segwit': True,
        'vsize': 105,
    },  # Bech32 P2WSH -- Not yet supported to sign (vsize corresponds to a 2-of-3 multisig input)
}


class Unspent:
    """Represents an unspent transaction output (UTXO)."""

    __slots__ = ('amount', 'confirmations', 'script', 'txid', 'txindex', 'type', 'vsize', 'segwit', 'sequence')

    def __init__(self, amount, confirmations, script, txid, txindex, type='p2pkh', vsize=None, segwit=None,
                 sequence=int.from_bytes(SEQUENCE, byteorder='little')):
        self.amount = amount
        self.confirmations = confirmations
        self.script = script
        self.txid = txid
        self.txindex = txindex
        self.type = type if type in UNSPENT_TYPES else 'unknown'
        self.vsize = vsize if vsize else UNSPENT_TYPES[self.type]['vsize']
        self.segwit = UNSPENT_TYPES[self.type]['segwit']
        self.sequence = sequence

    def to_dict(self):
        return {attr: getattr(self, attr) for attr in Unspent.__slots__}

    @classmethod
    def from_dict(cls, d):
        return Unspent(**{attr: d[attr] for attr in Unspent.__slots__})

    def __eq__(self, other):
        return (
            self.amount == other.amount
            and self.script == other.script
            and self.txid == other.txid
            and self.txindex == other.txindex
            and self.segwit == other.segwit
            and self.sequence == other.sequence
        )

    def __repr__(self):
        return 'Unspent(amount={}, confirmations={}, script={}, txid={}, txindex={}, segwit={}, sequence={})'.format(
            repr(self.amount),
            repr(self.confirmations),
            repr(self.script),
            repr(self.txid),
            repr(self.txindex),
            repr(self.segwit),
            repr(self.sequence)
        )

    def set_type(self, type, vsize=0):
        self.type = type if type in UNSPENT_TYPES else 'unknown'
        self.vsize = vsize if vsize else UNSPENT_TYPES[self.type]['vsize']
        self.segwit = UNSPENT_TYPES[self.type]['segwit']
        return self

    def opt_in_for_RBF(self):
        if self.sequence > 4294967293:
            self.sequence = 4294967293


import requests
from .exceptions import ApiError
from .convert import convert
from .utils.crypto import bytes_to_hex
from .tx import address_to_scriptpubkey
from .utils.meta import Unspent

class Blockchair:
    """retrieve blockchain data using the Blockchair API"""
    def __init__(self, network='mainnet'):
        network = network.lower()
        if network == 'mainnet':
            self.endpoint = "https://api.blockchair.com/bitcoin/"
        elif network == 'testnet':
            self.endpoint = "https://api.blockchair.com/bitcoin/testnet/"
        else:
            raise ValueError("Network unrecognized")

        self.address_info = "dashboards/address/{}"
        self.tx_info = "raw/transaction/{}"
        self.push_tx = "push/transaction"
        self.network = network

    def balance(self, address, unit='satoshi'):
        """get address balance"""
        endpoint = self.endpoint + self.address_info
        r = requests.get(endpoint.format(address))
        if r.status_code != 200:
            raise ApiError("Blockchair API unreachable")
        
        bal = r.json()['data'][address]['address']['balance']
        unit = unit.lower()
        if unit == 'satoshi':
            bal = int(bal)
        elif unit == 'btc':
            bal = convert.to_btc(bal)
        else:
            raise ValueError("Unit %s does not exist" % unit)

        return bal
    
    def txs(self, address):
        """get transaction hashes"""
        endpoint = self.endpoint + self.address_info
        params = {'offset': '0', 'limit': '1000'}
        r = requests.get(endpoint.format(address), params=params)
        if r.status_code != 200:
            raise ApiError("Blockchair API unreachable")

        transactions = r.json()['data'][address]['transactions']
        
        return transactions
    
    def utxo(self, address):
        """get unspents"""
        endpoint = self.endpoint + self.address_info
        params = {'offset': '0', 'limit': '1000'}
        r = requests.get(endpoint.format(address), params=params)
        if r.status_code != 200:
            raise ApiError("Blockchair API unreachable")

        res = r.json()

        block_height = res['context']['state']
        response = res['data'][address]
        script_pubkey = response['address']['script_hex']
        total_utxo = response['address']['unspent_output_count']
        utxos = response['utxo']
        _utxo = []
        
        for unspent in utxos:
            if unspent['block_id'] == -1:
                confs = 0
            else:
                confs = (block_height - unspent['block_id']) + 1
            _utxo.append({
                "value": unspent['value'],
                "confirmations": confs,
                "script": script_pubkey,
                "hash": unspent['transaction_hash'],
                "index": unspent['index']
                })

        return _utxo


class Blockstream:
    """retrieve blockchain data using the Blockstream API"""
    def __init__(self, network='mainnet'):
        network = network.lower()
        if network == "mainnet":
            self.endpoint = "https://blockstream.info/api/"
        elif network == "testnet":
            self.endpoint = "https://blockstream.info/testnet/api/"
        else:
            raise ValueError("Network unrecognized")
    
        self.address_info = "address/{}"
        self.tx_info = "tx/{}/hex"
        self.push_tx = "tx"
        self.unspent = "/utxo"
        self.network = network

    def balance(self, address, unit='satoshi'):
        """get address balance"""
        endpoint = self.endpoint + self.address_info
        r = requests.get(endpoint.format(address))
        if r.status_code != 200:
            raise ApiError("Blockstream API unreachable")
        
        r = r.json()
        funded = r['chain_stats']['funded_txo_sum'] + r['mempool_stats']['funded_txo_sum']
        spent = r['chain_stats']['spent_txo_sum'] + r['mempool_stats']['spent_txo_sum']

        bal = funded - spent

        unit = unit.lower()
        if unit == "satoshi":
            bal = int(bal)
        elif unit == "btc":
            bal = convert.to_btc(bal)
        else:
            raise ValueError("Unit %s does not exist" % unit)

        return bal

    def txs(self, address):
        """get transaction hashes"""
        endpoint = self.endpoint + self.address_info + '/txs/chain/{}'
        mempool_endpoint = self.endpoint + self.address_info + '/txs/mempool'

        transactions = []

        r = requests.get(mempool_endpoint.format(address))
        if r.status_code == 400:
            return []
        elif r.status_code != 200:
            raise ApiError("Blockstream API unreachable")

        response = r.json()
        unconfirmed = [tx['txid'] for tx in response]
        if len(unconfirmed) == 50:
            raise ApiError("Operation unsafe. Too many unconfirmed transactions.")

        r = requests.get(endpoint.format(address, ''))
        if r.status_code == 400:
            return []
        elif r.status_code != 200:
            raise ApiError("Blockstream API unreachable")

        response = r.json()

        total_txs = len(response)

        while total_txs > 0:
            transactions.extend(tx['txid'] for tx in response)

            response = requests.get(endpoint.format(address, transactions[-1])).json()
            total_txs = len(response)

        transactions.extend(unconfirmed)

        return transactions
    
    def utxo(self, address):
        """get unspents"""
        endpoint = self.endpoint + 'blocks/tip/height'
        r_block = requests.get(endpoint)
        if r_block.status_code != 200:
            raise ApiError("Blockstream API unreachable")

        block_height = int(r_block.text)

        utxo_endpoint = (self.endpoint+self.address_info+self.unspent).format(address)
        r = requests.get(utxo_endpoint)
        if r.status_code == 400 and r.text == "Too many history entries":
            raise ApiError("Excessive address")
        elif r.status_code != 200:
            raise ApiError("Blockstream API unreachable")

        script_pubkey = bytes_to_hex(address_to_scriptpubkey(address))
        return sorted(
            [
                Unspent(
                    tx["value"],
                    block_height - tx["status"]["block_height"] + 1 if tx["status"]["confirmed"] else 0,
                    script_pubkey,
                    tx["txid"],
                    tx["vout"],
                )
                for tx in r.json()
            ],
            key=lambda u: u.confirmations,
        )

    def broadcast(self, tx_hex):
        endpoint = self.endpoint + self.push_tx
        r = requests.post(endpoint, data=tx_hex)
        return True if r.status_code == 200 else False


from .key import Key
from .validate import validate
from .convert import convert
from .operation import operation
from .fees import get_fees
from .request import request_payment, parse_request
from .net import Net
from .rpc import AuthServiceProxy as Node
from .utils.opcodes import script
from .transaction import PrivateKey as Tx

name = "oogway"
__version__ = "0.6.1"
